/* WEBS: prototype simple and secure web server.  0.99 WORK IN PROGRESS
   960412

   "Secure" in this context does not refer to SSL or SHTTP.  This server
   hands out plain files from a [optional] chrooted area, and that's ALL.
   No scripts, no forms, no directories, no hex-translation, no nonsense.
   Reasonably comprehensive logging.  Wanna run scripts?  Bury redirects
   to alternate machines/ports/scripts in your HTML and take your chances.

   Runs under inetd, as root just long enough to render itself harmless.
   Then things get interesting: binds to a one-off port and hands out a
   redirect to same.  All subsequent interaction with a given caller uses
   the new port.  Queries from other sources are rejected and logged.
   After 10 minutes or so of inactivity, exits.  The caller will then have
   to come back in through inetd's listening port again.  Bummer.

   Nonroot users can build this without the CHROOT define and run their
   own web servers, constrained to a WEB_PATH tree of their own choosing.

   To build: fix up WEB_PATH, WEB_UID, etc below and compile with -DSYSTYPE
   as given in generic.h.  Link static if possible and fix up inetd.conf.

   later, maybe:
	have ifdeffed code to do that other stuff if *really* wanted?
	crypto-cookies: md5 time, host-magic, other stuff??
	perhaps a "secure" way to do scripts after all

   *Hobbit*, started 960115, up and limping 960125, almost exactly one
   year after avian.org came online... */

#include "generic.h"		/* comes with netcat, too... */

#include <stdio.h>
#include <string.h>		/* strcpy, strchr, yadda yadda */
#include <setjmp.h>		/* jmp_buf et al */
#include <fcntl.h>		/* O_whatever flags */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>		/* timeval, time_t */
#include <signal.h>
#include <syslog.h>
#include <errno.h>

#include <sys/socket.h>		/* basics, SO_ and AF_ defs, sockaddr, ... */
#include <netinet/in.h>		/* sockaddr_in, htons, in_addr */
#include <netinet/in_systm.h>	/* misc crud that netinet/ip.h references */
#include <netinet/ip.h>		/* IPOPT_LSRR, header stuff */
#include <netdb.h>		/* hostent, gethostby*, getservby* */
#include <arpa/inet.h>		/* inet_ntoa */

/* stuff specific to webs -- you will need to fiddle with this: */
#define WEB_PATH "/tmp/webs"	/* chroot [or prefixed] path */
#define WEB_IDX "index.html"	/* default index filename for "/" */
#define WEB_UID 69		/* something harmless on your system > 10 */
#define WEB_GID 69		/* likewise */
#define CHROOT 1		/* do the chroot-thing by default */
#define LOG_INITIAL 1		/* look up and log first connect */
#define FULL_REDIRECT 1		/* include orig req in redirect */
#define WEB_TMO 600		/* loop listen timeout [10 min] */
#define BSIZE 8192		/* max buffer size */

/* handy stuff swiped from netcat: */
#define SA struct sockaddr	/* socket overgeneralization braindeath */
#define SAI struct sockaddr_in	/* ... whoever came up with this model */
#define IA struct in_addr	/* ... should be taken out and shot, */
				/* ... not that TLI is any better.  sigh.. */
#define USHORT unsigned short	/* use these for options an' stuff */
#ifndef INADDR_NONE
#define INADDR_NONE 0xffffffff
#endif

/* globals: */
jmp_buf jbuf;			/* timer crud */
int jval = 0;			/* timer crud */
static char unknown[] = "(UNKNOWN)";
static char webpath[] = WEB_PATH;
static char webidx[] = WEB_IDX;
static char bbuf [BSIZE];	/* general-purpose big buffer */
static char fbuf [256];		/* filename buffer */
static struct stat dafile;
unsigned int bocount = 0;	/* how many bogus queries? */
unsigned int wtmo;		/* settable timeout */
SAI lclend;			/* local skt */
SAI remend;			/* remote skt */
char aclient [64];		/* original client ascii-addr */
IA oclient;			/* original client srcaddr */
FILE * chan;			/* the currently open net stream */
#ifdef SO_LINGER
struct linger lingerie;		/* we need *reliable* bra straps */
#endif

/* local-listener redirect.  We fill in "<addr>:<port>" etc on the fly */
static char redir[] =
  "HTTP/1.0 302 Port redirect [valid 15 seconds]\r\nLocation: http://";

#ifdef DEBUG
/* don't syslog for the mo'; send it over the original connection */
#undef syslog
#define syslog Dsyslog
void Dsyslog (junk, str, p1, p2, p3, p4, p5, p6)
  int junk;
  char * str;
  char * p1, * p2, * p3, * p4, * p5, * p6;
{
  fprintf (stderr, "LOG: ");
  fprintf (stderr, str, p1, p2, p3, p4, p5, p6);
  fprintf (stderr, "\n");
  fflush (stderr);
  sleep (1);
}
#endif /* DEBUG */

/* some code swiped from other things, mostly netcat */

/* bail :
   similar to that in netcat but closes "chan" and syslogs instead */
void bail (str, p1, p2, p3, p4, p5, p6)
  char * str;
  char * p1, * p2, * p3, * p4, * p5, * p6;
{
  register int x = errno;		/* save it */
/* if we have an open channel, error there and then nuke it */
  if (chan) {
    fprintf (chan, "oops! server failed\r\n");
    fflush (chan);
    sleep (1);
    (void) fclose (chan);
  }
  syslog (LOG_ERR, str, p1, p2, p3, p4, p5, p6);
  errno = x;
  if (errno)
    syslog (LOG_ERR, "%m");
  exit (1);
} /* bail */

/* timeout and other signal handling cruft */
void tmtravel ()
{
  signal (SIGALRM, SIG_IGN);
  alarm (0);
  if (jval == 0)
    bail ("spurious timer interrupt!");
  longjmp (jbuf, jval);
}

/* arm :
   set the timer.  Zero secs arg means unarm */
void arm (num, secs)
  unsigned int num;
  unsigned int secs;
{
  if (secs == 0) {			/* reset */
    signal (SIGALRM, SIG_IGN);
    alarm (0);
    jval = 0;
  } else {				/* set */
    signal (SIGALRM, tmtravel);
    alarm (secs);
    jval = num;
  } /* if secs */
} /* arm */

/* char conversion array.  See "chars.c".
   Valid: letters/numbers, plus " &-./:=?_~".  Anything else becomes "#" */
static unsigned char aflgs[] = {
    0,  35,  35,  35,  35,  35,  35,  35,	/* ^@ ^A ^B ^C ^D ^E ^F ^G  */
   35,  35,   0,  35,  35,   0,  35,  35,	/* ^H I^ ^J ^K ^L ^M ^N ^O  */
   35,  35,  35,  35,  35,  35,  35,  35,	/* ^P ^Q ^R ^S ^T ^U ^V ^W  */
   35,  35,  35,  35,  35,  35,  35,  35,	/* ^X ^Y ^Z ^[ ^\ ^] ^^ ^_  */
   32,  35,  35,  35,  35,  35,  38,  35,	/* sp  !  "  #  $  %  &  ,  */
   35,  35,  35,  35,  35,  45,  46,  47,	/*  (  )  *  +  ,  -  .  /  */
   48,  49,  50,  51,  52,  53,  54,  55,	/*  0  1  2  3  4  5  6  7  */
   56,  57,  58,  35,  35,  61,  35,  63,	/*  8  9  :  ;  <  =  >  ?  */
   35,  65,  66,  67,  68,  69,  70,  71,	/*  @  A  B  C  D  E  F  G  */
   72,  73,  74,  75,  76,  77,  78,  79,	/*  H  I  J  K  L  M  N  O  */
   80,  81,  82,  83,  84,  85,  86,  87,	/*  P  Q  R  S  T  U  V  W  */
   88,  89,  90,  35,  35,  35,  35,  95,	/*  X  Y  Z  [  \  ]  ^  _  */
   35,  97,  98,  99, 100, 101, 102, 103,	/*  `  a  b  c  d  e  f  g  */
  104, 105, 106, 107, 108, 109, 110, 111,	/*  h  i  j  k  l  m  n  o  */
  112, 113, 114, 115, 116, 117, 118, 119,	/*  p  q  r  s  t  u  v  w  */
  120, 121, 122,  35,  35,  35, 126,  35	/*  x  y  z  {  |  }  ~ ^?  */
}; /* aflgs */

/* fgetss :
   wrapper for fgets, that preconverts to a "safe" string and yanks off
   trailing newlines.  Doing all the work ourselves, instead of calling
   strchr/strlen/whatever, for security and speed. */
char * fgetss (buf, len, from)
  char * buf;
  size_t len;
  FILE * from;
{
  register int x;
  register char * p;
  register unsigned char q;
  char * bph;

  p = fgets (buf, len, from);		/* returns ptr to buf */
  if (! p)
    return (NULL);
  x = len - 1;
  buf[x] = '\0';			/* you're terminated, fucker */
  bph = p;
  for (x = 0; x < len; x++) {
    q = (unsigned char) (*p & 0x7f);	/* rip parity */
    q = aflgs[q];			/* convert thru array */
    *p = q;				/* stuff back into buffer */
    if (q == 0)
	return (bph);			/* self-terminating */
    p++;
  } /* for */
  return (NULL);			/* too long -- "can't happen" */
} /* fgetss */

/* now, support stuff specific to webs */

/* parseurl :
   Extract "path" part from various formats of query.  Constructs filename
   from first arg into second arg.  Disallows multiple dots and slashes.
   Only accepts an absolute path.  Supplies a default in place of "/".
   Returns pointer to final filename, or NULL if it's bogus. */
char * parseurl (str, out)
  char * str;
  char * out;
{
  register char *p, *q, *qs;
  register int x;
  register short s1, s2;	/* first/second found-space flags */

  if (strncmp (str, "GET ", 4))	/* only valid method here! */
    return (0);

  p = str + 4;			/* point after "GET " */
  if (*p != '/')
    return (0);			/* abspath, pleeze */
  out[0] = '\0';
  out[1] = '\0';
  q = out;

#ifndef CHROOT
/* If we're not chrooted, preload path-prefix to limit query space */
  strcpy (q, webpath);
  q += strlen (q);
#endif
  qs = q;			/* save start of supplied name */

/* copy filename chars from *p to *q, until we hit the second space or EOL.
   Suppresses any trailing "HTTP/x.y" as a side effect.  Rejects strings
   containing "//" or ".." or "/." or "./", all of which we consider
   questionable.  [thus you can't have filenames like ".foo" or "foo."]
   might put %dd hex-parsing in here someday, but not now... */
  for (x = 0; x < 250 ; x++) {
    switch (*p) {
	case  ' ':		/* next space, or */
	case '\0':		/* eol from fgetss */
	  *q = '\0';
	  break;
	case '.':
	case '/':
	  if (*(p+1) == '.')
	    return (0);
	  if (*(p+1) == '/')
	    return (0);
	default:		/* fall thru to normal char... */
	  *q = *p;
    } /* switch */
    if (! *q)
      break;			/* if no char copied, we're done */
    p++; q++;
  } /* for x */

/* special case for "GET /" -- substitute index filename */
  if (*qs == '/')
    if (*(qs+1) == '\0')
      strcat (out, webidx);
  return (out);
} /* parseurl */

/* bogusurl :
   hand back a stock obfuscatory answer to bogus queries, and log it.
   Side effect: closes "chan". */
void bogusurl (reason, spec)
  register char * reason;
  register char * spec;
{
  sleep (5);		/* make 'em wait for it */
  bocount++;
  fprintf (chan, "<title>Get real!</title><h2>Not found</h2>\n");
  fprintf (chan, "<p>That does not exist here.\n");
  if (bocount >= 3)
    fprintf (chan,
	"<p><b>Look, if you aren't gonna be serious, just GO AWAY.\n");
  fflush (chan);
  sleep (1);
  (void) fclose (chan);
  chan = (FILE *) 0;
  syslog (LOG_ERR, "%s strike %d: bogus %s: %.200s",
	aclient, bocount, reason, spec);
  if (bocount >= 3)
    exit (0);
}

/* handfile :
   retrieve a file and shovel it out a given fd.
   Side effect: closes "chan". */
void handfile (path, fd)
  char * path;
  int fd;
{
  register int x;
  register int y;
  int f;
  char * p;

  f = open (path, O_RDONLY);
  if (f < 0) {
    bogusurl ("open", path);
    return;
  }

#ifdef SO_LINGER
/* Set a linger time so our close doesn't interfere with any data still
   in flight.  I went through *hell* trying to figure out why browsers were
   complaining about aborted/reset connections.  I think this is why. */
  lingerie.l_onoff = 1;
  lingerie.l_linger = 10;		/* short, but enough */
  x = setsockopt (fd, SOL_SOCKET, SO_LINGER, (char *) &lingerie,
    sizeof (struct linger));
  if (x < 0)
    bail ("linger %d", fd);		/* fatal, for now */
#endif /* SO_LINGER */

/* Stock httpds use stdio all the way through.  I'll be damned if I'm gonna
   fool with that for raw data-shoveling, so here we go low-level. */
  errno = 0;
  while (1) {				/* no fancy select stuff here, */
    x = read (f, bbuf, BSIZE);		/* just a simple lockstep xfer */
    if (x <= 0)
      break;				/* whoops! */
    y = x;
    p = bbuf;
    while (x > 0) {
      y = write (fd, p, x);
      if (y <= 0)
	break;				/* whoops! */
      x -= y;
      p += y;
    } /* while x */
    if (x)
      break;				/* oops, couldn't write it all */
  } /* while 1 */
  y = errno;				/* save it */

#ifndef SO_LINGER
/* If we don't have linger-time, sleaze it.  This was proven to more or
   less work from the early testing... */
  sleep (1);
#endif /* SO_LINGER */

/* done; excrete the last bits and wipe up.  Have to close the raw fd first,
   so stdio doesn't come along and bugger it up somehow */
  (void) close (f);
  (void) close (fd);
  (void) fclose (chan);
  chan = (FILE *) 0;
  if (x == 0)
    syslog (LOG_INFO, "%s %.250s", aclient, path);
  else
    syslog (LOG_ERR, "%s %.250s I/O err %d", aclient, path, y);  /* nonfatal */
} /* handfile */

/* main :
   and away we go... */
main(argc, argv)
  int argc;
  char ** argv;
{
  register char * pp;
  register int x;
  int y;
  int netfd1;
  int netfd2;
  USHORT lp;
  struct hostent * hp = (struct hostent *) 0;

/* We just got started by inetd or some other handler.  Assume we're now
   running as root, talking to the caller via the usual stdin/stdout/stderr
   setup.  First insulate ourselves from randomness... */
#ifdef SIGURG
  signal (SIGURG, SIG_IGN);		/* no funny stuff from TCP */
#endif
#ifdef SIGPIPE
  signal (SIGPIPE, SIG_IGN);		/* write to aborted connections */
#endif

  chan = stderr;			/* from inetd */
  if (argc > 1)
    bail ("invocation");

/* get logging open while we can still see /dev/log */
  openlog ("webs", LOG_NDELAY, LOG_DAEMON);

/* okay, who iiiiiis this??  ...  >>> this is Jesus, Kent. <<< */
  memset (&remend, 0, sizeof (SA));
  y = sizeof (SAI);
  x = getpeername (0, (SA *) &remend, &y);
  if (x)
    bail ("peer");
  memcpy (&oclient, &remend.sin_addr, sizeof (IA));
  memset (aclient, 0, sizeof (aclient));
  sprintf (aclient, "[%.32s]", inet_ntoa (remend.sin_addr));

  memset (fbuf, 0, sizeof (fbuf));
  memset (bbuf, 0, BSIZE);

#ifdef LOG_INITIAL
  pp = bbuf;
  sprintf (bbuf, "%s ", aclient);
  pp += strlen (bbuf);
/* gawd, this egregiously clumsy host-lookup crap again.  Done before chroot
   so we can still find resolv.conf et al.  Disable for extra speed. */
  hp = gethostbyaddr ((char *) &oclient, sizeof (IA), AF_INET);
  if ((hp) && (hp->h_name))
    strncat (bbuf, hp->h_name, 128);
  else
    strcat (bbuf, unknown);
#endif

/* hmm, maybe we should first see if there *is* a web tree with an index */
  sprintf (fbuf, "%s/%s", webpath, webidx);
  x = stat (fbuf, &dafile);
  if (x)
    bail ("no web tree!!");

#ifndef DEBUG
/* render ourselves harmless */
#ifdef CHROOT
  x = chroot (webpath);
  if (x < 0)
    bail ("chroot web-path");
  x = chdir ("/");
  if (x < 0)
    bail ("chroot /");
#endif /* CHROOT */
  (void) setgid (WEB_GID);
  (void) setuid (WEB_UID);
#endif /* DEBUG */
  if (getuid() < 10)			/* Obie said he was just makin' sure */
    bail ("low uid");
  if (geteuid() < 10)
    bail ("low euid");

/* get a socket and make it available *before* we redirect */
  netfd1 = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (netfd1 < 0)
    bail ("socket");
  if (netfd1 < 3)			/* if we're missing 0/1/2, */
    bail ("skt lt 3");			/* something is VERY wrong */

/* get local-end addr and bind new random port only to that interface */
  memset (&lclend, 0, sizeof (SA));
  y = sizeof (SA);
  x = getsockname (0, (SA *) &lclend, &y);	/* get local-end addr */
  if (x)
    bail ("local getsock 0");
  lclend.sin_port = 0;
  x = bind (netfd1, (SA *) &lclend, sizeof (SA));	/* grab any port */
  if (x)
    bail ("local bind");
  y = sizeof (SA);
  x = getsockname (netfd1, (SA *) &lclend, &y);		/* find its number */
  if (x)
    bail ("local getsock nfd");
  lp = ntohs (lclend.sin_port);				/* and record it */
  if (lp == 0)
    bail ("local port");
/* phew!  Finally have our local port; give it a listen */
  x = listen (netfd1, 8);		/* really only 5 by default? */
  if (x)
    bail ("listen");
/* first timer is only 15 sec, in case they weren't serious */
  wtmo = 15;

#ifdef LOG_INITIAL
  syslog (LOG_INFO, "%s given port %d", bbuf, lp);
#endif

/* see if they *do* send anything in, to ward off crap like port scans */
  errno = 0;
  pp = (char *) 0;
  x = sizeof (fbuf) - sizeof (webpath) - 2;	/* limit input length... */
  arm (1, wtmo);
  if (setjmp (jbuf) == 0) {
    pp = fgetss (bbuf, x, stdin);
  }
  arm (0, 0);
  if (! pp)
    bail ("fake hit");

#ifdef FULL_REDIRECT
/* Find the request-path and include it in the redirect */
  pp = parseurl (bbuf, fbuf);
  if (! pp) {
    bogusurl ("first hit", bbuf);
    exit (0);				/* for inetd */
  }
#ifndef CHROOT
/* fbuf comes back with full web-tree-path.  Skip our private bits ... */
  pp = fbuf + strlen (webpath);		/* "/u/web/blah" -> "/blah" */
#endif /* CHROOT */
#else /* FULL_REDIRECT */
  fbuf[0] = '/';
  fbuf[1] = '\0';
  pp = fbuf;
#endif /* FULL_REDIRECT */

/* hand out the redirect, and clean up */
  printf ("%s%s:%d%s\r\n", redir, inet_ntoa (lclend.sin_addr), lp, pp);
  fflush (stdout);
#ifndef DEBUG			/* leave stderr open for fake-syslog */
  (void) fclose (chan);
  chan = (FILE *) 0;
  (void) close (2);
#endif
  (void) close (1);
  (void) close (0);		/* inetd connection dies here! */

/* RUNNING LOOP: set timer and wait for a connection */
  while (1) {
    y = sizeof (SA);
    arm (2, wtmo);
    if (setjmp (jbuf) == 0) {
      x = accept (netfd1, (SA *) &remend, &y);
    } else {
      break;
    } /* setjmp */
    arm (0, 0);
    if (x < 0)
      bail ("accept");
    netfd2 = x;
    wtmo = WEB_TMO;		/* normal timeout for subsequent queries */

/* new connection; assign channel to it and disable buffering.
   All this just so we can do fprintf, ferkrissake... */
    chan = fdopen (netfd2, "r+");
    if (chan == 0)
      bail ("new chan for netfd2");
    setvbuf (chan, (char *) 0, _IONBF, 0);	/* pain in the ass stdio */

/* suck in the request, limiting length to what we can safely handle */
    pp = (char *) 0;
    x = sizeof (fbuf) - sizeof (webpath) - 2;
    arm (3, wtmo);
    if (setjmp (jbuf) == 0) {
      pp = fgetss (bbuf, x, chan);
    }
    arm (0, 0);
    if (! pp) {
      bogusurl ("input", bbuf);
      continue;
    }

/* is this our original client??  If not, be mightily hostile. */
    y = sizeof (SA);
    x = getpeername (netfd2, (SA *) &remend, &y);
    if (x)
      bail ("new peer");
    if (memcmp (&oclient, &remend.sin_addr, sizeof (IA))) {
      sleep (5);
      fprintf (chan,
	"Who the fuck are you?  Take a hike.\n");
      fflush (chan);
      sleep (1);
      (void) fclose (chan);
      chan = (FILE *) 0;
      (void) close (netfd2);
      syslog (LOG_ERR, "Denied port %d hit from [%s]: %.128s",
	lp, inet_ntoa (remend.sin_addr), bbuf);
      continue;
    } /* memcmp client[s] */

/* finally, try to handle the damn thing, and loop.
   Both bogusurl and handfile flush and close "chan" for us */
    pp = parseurl (bbuf, fbuf);
    if (! pp) {
      bogusurl ("parse", bbuf);
      continue;
    }
    x = stat (pp, &dafile);		/* not lstat, so symlinks work */
    if (x) {
      bogusurl ("file", pp);
      continue;
    }
    x = 1;
    if ((dafile.st_mode & S_IFMT) == S_IFREG)
      handfile (pp, netfd2);
    else
      bogusurl ("filetype", pp);
  } /* while 1 */

/* here when timeout fired; log it and die */
  (void) close (netfd2);
  (void) close (netfd1);
  syslog (LOG_INFO, "%s %d timeout", aclient, lp);
  exit (0);
} /* main */

